/* threads.c
** - Thread Abstraction Functions
**
** Copyright (c) 1999, 2000 the icecast team
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any latfer version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>

#ifndef __USE_MISC
# define __USE_MISC
#endif

#ifndef __USE_GNU
# define __USE_GNU
#endif

#ifndef __USE_BSD
# define __USE_BSD
#endif

#ifndef __EXTENSIONS__
# define __EXTENSIONS__
#endif

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif

#ifndef _THREAD_SAFE
# define _THREAD_SAFE
#endif

#ifndef _REENTRANT
# define _REENTRANT
#endif

#ifndef __USE_POSIX
# define __USE_POSIX
#endif

#ifndef __STDC__
# define __STDC__
#endif

/* This for freebsd (needed on 3.2 at least) */
#ifdef SOMEBSD
# ifndef _POSIX_VERSION
# define _POSIX_VERSION 199309L
# endif
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef _WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

#include <signal.h>

#include "thread.h"

#include "avl.h"

/* INTERNAL DATA */

#ifdef _WIN32
# define STACKSIZE 8192
#else
# define STACKSIZE 1024 * 250
#endif

/* log types */
#define LOGMUTEX "mutex"
#define LOGTHREAD "thread"
#define LOGWARNING "warning"
#define LOGERROR "error"
#define LOGFATAL "fatal"

/* thread starting structure */
typedef struct thread_start_tag {
	/* the real start routine and arg */
	void *(*start_routine)(void *);
	void *arg;

	/* the other stuff we need to make sure this thread is inserted into
	** the thread tree
	*/
	thread_t *thread;
	sys_thread_t sys_thread;
} thread_start_t;

static long _next_thread_id = 0;
static int _thread_initialized = 0;
static avl_tree *_threadtree = NULL;
static mutex_t _threadtree_mutex = { -1, NULL, MUTEX_STATE_UNINIT, NULL, -1 };

static long _next_mutex_id = 0;
static avl_tree *_mutextree = NULL;
static mutex_t _mutextree_mutex = { -1, NULL, MUTEX_STATE_UNINIT, NULL, -1 };

static mutex_t _library_mutex = { -1, NULL, MUTEX_STATE_UNINIT, NULL, -1 };
static mutex_t _multi_mutex = { -1, NULL, MUTEX_STATE_UNINIT, NULL, -1 }; 

#ifndef _WIN32
static pthread_attr_t _pthread_defaultattr;
#endif

/* INTERNAL FUNCTIONS */

/* avl tree functions */
static int _compare_mutexes(void *compare_arg, void *a, void *b);
static int _compare_threads(void *compare_arg, void *a, void *b);
static int _free_mutex(void *key);
static int _free_thread(void *key);

/* mutex fuctions */
static void _create_mutex(mutex_t *mutex);
static void _mutex_lock(mutex_t *mutex);
static void _unlock_mutex(mutex_t *mutex);

/* misc thread stuff */
static sys_thread_t _get_sys_thread();
static void *_start_routine(void *arg);
static void _thread_catch_signals ();
static void _thread_setup_default_attributes();
static void _thread_block_signals ();

/* logging */
void thread_log(char *type, int level, char *fmt, ...);

/* LIBRARY INITIALIZATION */

void thread_initialize()
{
	thread_t *thread;

	/* create all the interal mutexes, and initialize the mutex tree */

	_thread_setup_default_attributes();

	_mutextree = avl_tree_new (_compare_mutexes, NULL);

	/* we have to create this one by hand, because there's no
	** mutextree_mutex to lock yet! 
	*/
	_create_mutex(&_mutextree_mutex);
	_mutextree_mutex.mutex_id = _next_mutex_id++;
	avl_insert (_mutextree, (void *)&_mutextree_mutex);

	thread_create_mutex(&_threadtree_mutex);
	thread_create_mutex(&_library_mutex);	
	thread_create_mutex(&_multi_mutex);

	/* initialize the thread tree and insert the main thread */

	_threadtree = avl_tree_new(_compare_threads, NULL);

	thread = (thread_t *)malloc(sizeof(thread_t));

	thread->thread_id = _next_thread_id++;
	thread->line = 0;
	thread->file = strdup("main.c");
	thread->sys_thread = _get_sys_thread();
	thread->create_time = time(NULL);
	thread->name = strdup("Main Thread");

	avl_insert (_threadtree, (void *)thread);

	_thread_catch_signals ();

	_thread_initialized = 1;
}

void thread_shutdown()
{
	if (_thread_initialized == 1) {
		thread_destroy_mutex (&_multi_mutex);
		thread_destroy_mutex (&_library_mutex);
		thread_destroy_mutex(&_threadtree_mutex);
		thread_destroy_mutex(&_mutextree_mutex);
		
		avl_tree_free (_mutextree, _free_mutex);
		avl_tree_free (_threadtree, _free_thread);
	}
}

int
thread_is_initialized () 
{
	return _thread_initialized;
}

/*
 * Signals should be handled by the main thread, nowhere else.
 * I'm using POSIX signal interface here, until someone tells me
 * that I should use signal/sigset instead
 */
static void
_thread_block_signals ()
{
#if defined(HAVE_SIGACTION) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t ss;

        sigfillset (&ss);

        /* These ones we want */
        sigdelset (&ss, SIGKILL);
        sigdelset (&ss, SIGSTOP);
        sigdelset (&ss, SIGTERM);
        sigdelset (&ss, SIGSEGV);
        sigdelset (&ss, SIGBUS);
        if (pthread_sigmask (SIG_BLOCK, &ss, NULL) != 0)
                _thread_log (LOGERROR, 0, "Pthread_sigmask() failed for blocking signals");
#endif
}

/*
 * Let the calling thread catch all the relevant signals
 */
static void
_thread_catch_signals ()
{
#if defined(HAVE_SIGACTION) && defined(HAVE_PTHREAD_SIGMASK)
        sigset_t ss;

        sigemptyset (&ss);

        /* These ones should only be accepted by the signal handling thread (main thread) */
        sigaddset (&ss, SIGHUP);
        sigaddset (&ss, SIGCHLD);
        sigaddset (&ss, SIGINT);
#ifdef SIGPIPE
        sigaddset (&ss, SIGPIPE);
#endif

        if (pthread_sigmask (SIG_UNBLOCK, &ss, NULL) != 0)
                _thread_log(LOGERROR, 0, "WARNING: pthread_sigmask() failed!");
#endif
}


long thread_create_c(char *name, void *(*start_routine)(void *), void *arg, int line, char *file)
{
#ifndef _WIN32
	pthread_attr_t attr;
	int i;
#else
	HANDLE ret;
#endif

	thread_t *thread = (thread_t *)malloc(sizeof(thread_t));
	thread_start_t *start = (thread_start_t *)malloc(sizeof(thread_start_t));

	thread->line = line;
	thread->file = strdup(file);

	_mutex_lock(&_threadtree_mutex);	
	thread->thread_id = _next_thread_id++;
	_unlock_mutex(&_threadtree_mutex);

	thread->name = strdup(name);
	thread->create_time = time(NULL);

	start->start_routine = start_routine;
	start->arg = arg;
	start->thread = thread;

#ifdef _WIN32
	ret = CreateThread(NULL, STACKSIZE, (LPTHREAD_START_ROUTINE)_start_routine, start->arg, 0, &thread->sys_thread);
#else
	pthread_attr_init(&attr);

# ifdef HAVE_PTHREAD_ATTR_SETSTACKSIZE
	pthread_attr_setstacksize(&attr, STACKSIZE);
# endif

	/* Why is this here?  Should we have to keep trying? -jm 
	   No reason other than that if the system is temporarily out of
	   resources, then icecast survives. - Eel   */
	for (i = 0; i < 10; i++) {
# ifdef HPUX
		if (pthread_create((pthread_t *)&thread->sys_thread, pthread_attr_default, (pthread_startroutine_t)_start_routine, (pthread_addr_t)start) == 0)
# else
	        if (pthread_create(&thread->sys_thread, &attr, _start_routine, start) == 0)
# endif
			break;
		else
			thread_log(LOGERROR, 0, "Could not create new thread, retrying");
	}

	pthread_attr_destroy(&attr);	
#endif

#ifdef _WIN32
	if (ret == NULL) {
#else
	if (i >= 10) {
#endif
		thread_log(LOGFATAL, 0, "System won't let me create more threads, giving up");
		return -1;
	}

	return thread->thread_id;
}

/* _create_mutex 
** 
** creates a mutex
*/
static void _create_mutex(mutex_t *mutex)
{
	mutex->thread_id = MUTEX_STATE_NEVERLOCKED;
	mutex->line = -1;

#ifdef _WIN32
	InitializeCriticalSection(&mutex->sys_mutex);
#else
# ifdef HPUX
	pthread_mutex_init(&mutex->sys_mutex, pthread_mutexattr_default);
# else
	pthread_mutex_init(&mutex->sys_mutex, NULL);
# endif
#endif
}

void thread_create_mutex_c(mutex_t *mutex, int line, char *file)
{
	_create_mutex(mutex);

#ifdef DEBUG_MUTEXES
	_mutex_lock(&_mutextree_mutex);
	mutex->mutex_id = _next_mutex_id++;
	avl_insert (_mutextree, (void *)mutex);
	_unlock_mutex(&_mutextree_mutex);
#endif
}

void thread_destroy_mutex (mutex_t *mutex)
{
#ifdef _WIN32
	DeleteCriticalSection(&mutex->sys_mutex);
#else
	pthread_mutex_destroy(&mutex->sys_mutex);
#endif

#ifdef DEBUG_MUTEXES
	_mutex_lock(&_mutextree_mutex);
	avl_delete (_mutextree, mutex, _free_mutex);
	_unlock_mutex(&_mutextree_mutex);
#endif
}

void thread_mutex_lock_c(mutex_t *mutex, int line, char *file)
{
#ifdef DEBUG_MUTEXES
	thread_t *th = thread_self();

	if (!th) {
		thread_log(LOGWARNING, 0, "No mt record for %u in lock [%s:%d]", thread_self(), file, line);
	}

	thread_log(LOGMUTEX, 3, "Locking %p (%s) on line %d in file %s by thread %d", mutex, mutex->name, line, file, th ? th->thread_id : -1);

# ifdef CHECK_MUTEXES
	/* Just a little sanity checking to make sure that we're locking
	** mutexes correctly
	*/

	if (th) {
		int locks = 0;
		avl_node *node;
		mutex_t *tmutex;

		_mutex_lock(&_mutextree_mutex);

		node = avl_get_first (_mutextree);
		
		while (node) {
			tmutex = (mutex_t *)node->key;

			if (tmutex->mutex_id == mutex->mutex_id) {
				if (tmutex->thread_id == th->thread_id) { 
					/* Deadlock, same thread can't lock the same mutex twice */
					thread_log(LOGERROR, 0, "DEADLOCK AVOIDED (%d == %d) on mutex [%s] in file %s line %d by thread %d [%s]", 
					     tmutex->thread_id, th->thread_id, mutex->name ? mutex->name : "undefined", file, line, th->thread_id, th->name);

					_unlock_mutex(&_mutextree_mutex);
					return;
				}
			} else if (tmutex->thread_id == th->thread_id) { 
				/* Mutex locked by this thread (not this mutex) */
				locks++;
			}

			node = avl_get_next(node);
		}

		if (locks > 0) { 
			/* Has already got a mutex locked */
			if (_multi_mutex.thread_id != th->thread_id) {
				/* Tries to lock two mutexes, but has not got the double mutex, norty boy! */
				thread_log(LOGWARNING, 1, "(%d != %d) Thread %d [%s] tries to lock a second mutex [%s] in file %s line %d, without locking double mutex!",
				     _multi_mutex.thread_id, th->thread_id, th->thread_id, th->name, mutex->name ? mutex->name : "undefined", file, line);
			}
		}
		
		_unlock_mutex(&_mutextree_mutex);
	}
# endif /* CHECK_MUTEXES */
	
	_mutex_lock(mutex);
	
	_mutex_lock(&_mutextree_mutex);

	thread_log(LOGMUTEX, 3, "Locked %p by thread %d", mutex, th ? th->thread_id : -1);
	mutex->line = line;
	if (th) {
		mutex->thread_id = th->thread_id;
	}

	_unlock_mutex(&_mutextree_mutex);
#else
	_mutex_lock(mutex);
#endif /* DEBUG_MUTEXES */
}

void thread_mutex_unlock_c(mutex_t *mutex, int line, char *file)
{
#ifdef DEBUG_MUTEXES
	thread_t *th = thread_self();

	if (!th) {
		thread_log(LOGERROR, 0, "No record for %u in unlock [%s:%d]", thread_self(), file, line);
	}

	thread_log(LOGMUTEX, 3, "Unlocking %p (%s) on line %d in file %s by thread %d", mutex, mutex->name, line, file, th ? th->thread_id : -1);

	mutex->line = line;

# ifdef CHECK_MUTEXES
	if (th) {
		int locks = 0;
		avl_node *node;
		mutex_t *tmutex;

		_mutex_lock(&_mutextree_mutex);

		while (node) {
			tmutex = (mutex_t *)node->key;

			if (tmutex->mutex_id == mutex->mutex_id) {
				if (tmutex->thread_id != th->thread_id) {
					thread_log(LOGERROR, 0, "ILLEGAL UNLOCK (%d != %d) on mutex [%s] in file %s line %d by thread %d [%s]", tmutex->thread_id, th->thread_id, 
					     mutex->name ? mutex->name : "undefined", file, line, th->thread_id, th->name);
					_unlock_mutex(&_mutextree_mutex);
					return;
				}
			} else if (tmutex->thread_id == th->thread_id) {
				locks++;
			}

			node = avl_get_next (node);
		}

		if ((locks > 0) && (_multi_mutex.thread_id != th->thread_id)) {
			/* Don't have double mutex, has more than this mutex left */
		
			thread_log(LOGWARNING, 0, "(%d != %d) Thread %d [%s] tries to unlock a mutex [%s] in file %s line %d, without owning double mutex!",
			     _multi_mutex.thread_id, th->thread_id, th->thread_id, th->name, mutex->name ? mutex->name : "undefined", file, line);
		}

		_unlock_mutex(&_mutextree_mutex);
	}
# endif  /* CHECK_MUTEXES */

	_unlock_mutex(mutex);

	_mutex_lock(&_mutextree_mutex);

	thread_log(LOGMUTEX, 3, "Unlocked %p by thread %d", mutex, th ? th->thread_id : -1);
	mutex->line = -1;
	if (mutex->thread_id == th->thread_id) {
		mutex->thread_id = MUTEX_STATE_NOTLOCKED;
	}

	_unlock_mutex(&_mutextree_mutex);
#else
	_unlock_mutex(mutex);
#endif /* DEBUG_MUTEXES */
}

void thread_lock_multimutex_c(int line, char *file)
{
	thread_mutex_lock_c(&_multi_mutex, line, file);
}

void thread_unlock_multimutex_c(int line, char *file)
{
	thread_mutex_unlock_c(&_multi_mutex, line, file);
}

void thread_exit_c(int val, int line, char *file)
{
	thread_t *th = thread_self();

#if defined(DEBUG_MUTEXES) && defined(CHECK_MUTEXES)
	if (th) {
		avl_node *node;
		mutex_t *tmutex;
		char name[40];

		_mutex_lock(&_mutextree_mutex);

		while (node) {
			tmutex = (mutex_t *)node->key;

			if (tmutex->thread_id == th->thread_id) {
				thread_log(LOGWARNING, 0, "Thread %d [%s] exiting in file %s line %d, without unlocking mutex [%s]", 
				     th->thread_id, th->name, file, line, mutex_to_string(tmutex, name));
			}

			node = avl_get_next (node);
		}

		_unlock_mutex(&_mutextree_mutex);
	}
#endif
	
	if (th)	{
		thread_log(LOGTHREAD, 1, "Removing thread %d [%s] started at [%s:%d], reason: 'Thread Exited'", th->thread_id, th->name, th->file, th->line);

		_mutex_lock(&_threadtree_mutex);
		avl_delete (_threadtree, th, _free_thread);
		_unlock_mutex(&_threadtree_mutex);
	}
	
#ifdef _WIN32
	ExitThread(val);
#else
	pthread_exit((void *)val);
#endif
}

void thread_sleep(unsigned long len)
{
#ifdef _WIN32
	Sleep(len / 1000);
#else
# ifdef HAVE_NANOSLEEP
	struct timespec time_sleep;
	struct timespec time_remaining;
	int ret;

	time_sleep.tv_sec = len / 1000000;
	time_sleep.tv_nsec = len % 1000000;

	ret = nanosleep(&time_sleep, &time_remaining);
	while (ret != 0 && errno == EINTR) {
		time_sleep.tv_sec = time_remaining.tv_sec;
		time_sleep.tv_nsec = time_remaining.tv_nsec;
		
		ret = nanosleep(&time_sleep, &time_remaining);
	}
# else
	struct timeval tv;

	tv.tv_sec = len / 1000000;
	tv.tv_usec = (len % 1000000) / 1000;

	select(0, NULL, NULL, NULL, &tv);
# endif
#endif
}

static void *_start_routine(void *arg)
{
	thread_start_t *start = (thread_start_t *)arg;
	void *(*start_routine)(void *) = start->start_routine;
	void *real_arg = start->arg;
	thread_t *thread = start->thread;

	_thread_block_signals ();

	free(start);

	/* insert thread into thread tree here */
	_mutex_lock(&_threadtree_mutex);
	thread->sys_thread = _get_sys_thread();
	avl_insert (_threadtree, (void *)thread);
	_unlock_mutex(&_threadtree_mutex);

	thread_log(LOGTHREAD, 1, "Added thread %d [%s] started at [%s:%d]", thread->thread_id, thread->name, thread->file, thread->line);

#ifndef _WIN32
	pthread_detach(thread->sys_thread);
#endif

#ifdef PTHREAD_CANCEL_ENABLE
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
#endif

	/* call the real start_routine and start the thread
	** this should never exit!
	*/
	(start_routine)(real_arg);

	thread_log(LOGWARNING, 0, "Thread x should never exit from here!!!");

	return NULL;
}

static sys_thread_t _get_sys_thread()
{
#ifdef _WIN32
    return GetCurrentThreadId();
#else
    return pthread_self();
#endif
}

static int _equal(sys_thread_t st1, sys_thread_t st2)
{
#ifdef _WIN32
    return (st1 == st2) ? 1 : 0;
#else
    return pthread_equal(st1, st2);
#endif
}

thread_t *thread_self()
{
	avl_node *node;
	thread_t *th;
	sys_thread_t sys_thread = _get_sys_thread();

	_mutex_lock(&_threadtree_mutex);

	if (_threadtree == NULL) {
		thread_log(LOGERROR, 0, "Thread tree is empty, this must be wrong!");
		_unlock_mutex(&_threadtree_mutex);
		return NULL;
	}
	
	node = avl_get_first (_threadtree);
	
	while (node) {
		th = (thread_t *)node->key;

		if (th && _equal(sys_thread, th->sys_thread)) {
			_unlock_mutex(&_threadtree_mutex);
			return th;
		}
		
		node = avl_get_next (node);
	}
	_unlock_mutex(&_threadtree_mutex);


	thread_log(LOGERROR, 0, "Nonexistant thread alive...");
	
	return NULL;
}

void thread_rename(const char *name)
{
	thread_t *th;

	th = thread_self();
	if (th->name) {
		free(th->name);
	}

	th->name = strdup(name);
}

static void _mutex_lock(mutex_t *mutex) 
{
#ifdef _WIN32
	EnterCriticalSection(&mutex->sys_mutex);
#else
	pthread_mutex_lock(&mutex->sys_mutex);
#endif
}

static void _unlock_mutex(mutex_t *mutex)
{
#ifdef _WIN32
	LeaveCriticalSection(&mutex->sys_mutex);
#else
	pthread_mutex_unlock(&mutex->sys_mutex);
#endif
}


void thread_library_lock()
{
	_mutex_lock(&_library_mutex);
}

void thread_library_unlock()
{
	_unlock_mutex(&_library_mutex);
}

static void _thread_setup_default_attributes()
{
#if !defined(_WIN32) && defined(PTHREAD_CREATE_DETACHED)
	pthread_attr_init(&_pthread_defaultattr);
	pthread_attr_setdetachstate(&_pthread_defaultattr, PTHREAD_CREATE_DETACHED);
#endif
}

/* AVL tree functions */

static int _compare_mutexes(void *compare_arg, void *a, void *b)
{
	mutex_t *m1, *m2;

	m1 = (mutex_t *)a;
	m2 = (mutex_t *)b;

	if (m1->mutex_id > m2->mutex_id)
		return 1;
	if (m1->mutex_id < m2->mutex_id)
		return -1;
	return 0;
}

static int _compare_threads(void *compare_arg, void *a, void *b)
{
	thread_t *t1, *t2;

	t1 = (thread_t *)a;
	t2 = (thread_t *)b;

	if (t1->thread_id > t2->thread_id)
		return 1;
	if (t1->thread_id < t2->thread_id)
		return -1;
	return 0;
}

static int _free_mutex(void *key)
{
	mutex_t *m;

	m = (mutex_t *)key;

	if (m && m->file) {
		free(m->file);
		m->file = NULL;
	}

	/* Can't free here, cause we didn't allocate it in the first place 
	if (m)
	free(m); */

	return 1;
}

static int _free_thread(void *key)
{
	thread_t *t;

	t = (thread_t *)key;

	if (t->file)
		free(t->file);
	if (t->name)
		free(t->name);

	free(t);

	return 1;
}








